/*
 * Copyright (c) 2016 Intel Corporation.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr.h>
#include <drivers/flash.h>
#include <device.h>
#include <devicetree.h>
#include <stdio.h>
#include <string.h>
#include <kernel.h>
#include <helpers/nrfx_gppi.h>
#include <nrfx_dppi.h>
#include <nrfx_timer.h>
#include <logging/log.h>
LOG_MODULE_REGISTER(spi_flash, LOG_LEVEL_INF);

nrfx_timer_t timer1 = NRFX_TIMER_INSTANCE(1);
nrfx_timer_t timer2 = NRFX_TIMER_INSTANCE(2);
#if (CONFIG_SPI_NOR - 0) ||				\
	DT_NODE_HAS_STATUS(DT_INST(0, jedec_spi_nor), okay)
#define FLASH_DEVICE DT_LABEL(DT_INST(0, jedec_spi_nor))
#define FLASH_NAME "JEDEC SPI-NOR"
#elif (CONFIG_NORDIC_QSPI_NOR - 0) || \
	DT_NODE_HAS_STATUS(DT_INST(0, nordic_qspi_nor), okay)
#define FLASH_DEVICE DT_LABEL(DT_INST(0, nordic_qspi_nor))
#define FLASH_NAME "JEDEC QSPI-NOR"
#elif DT_NODE_HAS_STATUS(DT_INST(0, st_stm32_qspi_nor), okay)
#define FLASH_DEVICE DT_LABEL(DT_INST(0, st_stm32_qspi_nor))
#define FLASH_NAME "JEDEC QSPI-NOR"
#else
#error Unsupported flash driver
#endif

#if defined(CONFIG_BOARD_ADAFRUIT_FEATHER_STM32F405)
#define FLASH_TEST_REGION_OFFSET 0xf000
#elif defined(CONFIG_BOARD_ARTY_A7_ARM_DESIGNSTART_M1) || \
	defined(CONFIG_BOARD_ARTY_A7_ARM_DESIGNSTART_M3)
/* The FPGA bitstream is stored in the lower 536 sectors of the flash. */
#define FLASH_TEST_REGION_OFFSET \
	DT_REG_SIZE(DT_NODE_BY_FIXED_PARTITION_LABEL(fpga_bitstream))
#else
#define FLASH_TEST_REGION_OFFSET 0x00000
#endif

// inhirit from Henrik's flash structure setup 
#define FLASH_BYTES_PER_PAGE 256
#define FLASH_PAGES_PER_SECTOR 16
#define FLASH_BYTES_PER_SECTOR (FLASH_BYTES_PER_PAGE*FLASH_PAGES_PER_SECTOR)
#define FLASH_CHIP_DENSITY	(0x17 /* 23 */)
#define FLASH_NUM_PAGES (1U << (FLASH_CHIP_DENSITY-8))
#define FLASH_PAGE_MASK (FLASH_NUM_PAGES - 1)
#define FLASH_FSYS_SECTOR 0 /* The sector number of the file system sector (containing file pointers) */
#define FLASH_KVAL_SECTOR 1 /* The sector number of the special key-value sector */
#define FLASH_BULK_SECTOR 2 /* The first sector of the space dedicated to files */

#define FLASH_FSYS_START_PAGE (FLASH_FSYS_SECTOR*FLASH_PAGES_PER_SECTOR)
#define FLASH_KVAL_START_PAGE (FLASH_KVAL_SECTOR*FLASH_PAGES_PER_SECTOR)
#define FLASH_BULK_START_PAGE (FLASH_BULK_SECTOR*FLASH_PAGES_PER_SECTOR)

#define FLASH_FSYS_START_ADDRESS (FLASH_FSYS_SECTOR*FLASH_BYTES_PER_SECTOR)
#define FLASH_KVAL_START_ADDRESS (FLASH_KVAL_SECTOR*FLASH_BYTES_PER_SECTOR)
#define FLASH_BULK_START_ADDRESS (FLASH_BULK_SECTOR*FLASH_BYTES_PER_SECTOR)

// if sampling time is too quick ex 1ms, the write_flash will delay the process, and the next page will start from 141 instead of 127

#define SAMPLE_TIME_ms 		100  // have tried 10us sampling with 100 pages, all good
#define READ_TIME_ms 		10 
#define ANALOG_NUM_BUF_pages 2  // no problem to set 100 pages of buffer


uint8_t outputbuf[ANALOG_NUM_BUF_pages << 8]; 
uint8_t outputbuf1[256];
uint16_t count;

int st,ed;

const struct device *flash_dev;

K_FIFO_DEFINE(flash_write_fifo);
struct data_item_t {
    void *fifo_reserved;   /* 1st word reserved for use by fifo */
    uint16_t counter;
};


void timer1_evt_handler(nrf_timer_event_t event_type, void * p_context)
{	
	switch(event_type)
	{	
		case NRF_TIMER_EVENT_COMPARE0: ;
			struct data_item_t tx_data = { .counter=++count };
			k_fifo_put(&flash_write_fifo, &tx_data);
			printk("count: %d\n",count);
			break;
		
		default:
			//Do nothing.
			break;
	}    
}

void timer_init(void)
{
	nrfx_err_t err = NRFX_SUCCESS;
	nrfx_timer_config_t timer_cfg = NRFX_TIMER_DEFAULT_CONFIG;
	timer_cfg.bit_width = NRF_TIMER_BIT_WIDTH_32;

	err = nrfx_timer_init(&timer1, &timer_cfg, timer1_evt_handler);
	if(err != NRFX_SUCCESS)
	{
		LOG_ERR("timer 1 Error! Could not initialize TIMER1: %d\n", err);
	}
	uint32_t time_ticks = nrfx_timer_ms_to_ticks(&timer1, SAMPLE_TIME_ms);
	nrfx_timer_extended_compare(&timer1, NRF_TIMER_CC_CHANNEL0, time_ticks, NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK, true);	
	nrfx_timer_enable(&timer1);
	IRQ_DIRECT_CONNECT(TIMER1_IRQn, 0, nrfx_timer_1_irq_handler, 0);
	
}
void timer_deinit(void){
	nrfx_timer_disable(&timer1);
	irq_disable(TIMER1_IRQn);
}
// static inline bool caughtUpWithWriting() {
// 	return page_address_read >= page_address_write; //&&
// 		   //(!filled_flash || page_address_read < starting_page);
// }

static void flash_write_thread(void *a1, void *a2, void *a3 ){
	size_t siz, wbsiz;
	int rc;
	uint32_t st,ed;
	
	const struct device *flash_dev;
	
	printf("\n" FLASH_NAME " SPI flash testing\n");
	printf("==========================\n");

	flash_dev = device_get_binding(FLASH_DEVICE);

	if (!flash_dev) {
		printf("SPI flash driver %s was not found!\n",
		       FLASH_DEVICE);
		return;
	}
	
	siz=flash_get_page_count(flash_dev);
	wbsiz=flash_get_write_block_size(flash_dev);

	
   // erase is needed before write, otherwise there may be error between write and read
	printf("\nTest 1: Flash erase\n");
	st=k_cycle_get_32();
	rc = flash_erase(flash_dev, FLASH_TEST_REGION_OFFSET,
			 (ANALOG_NUM_BUF_pages/FLASH_PAGES_PER_SECTOR+1)*FLASH_BYTES_PER_SECTOR); // erase take integer sectors of bytes
	ed=k_cycle_get_32();
	if (rc != 0) {
		printf("Flash erase failed! %d\n", rc);
	} else {
		printf("Flash erase succeeded! %d , %d \n",st,ed);
	}

	struct data_item_t  *rx_data;
	uint8_t tembuf[FLASH_BYTES_PER_PAGE];
	int reads, page_address_write;
	reads=0;
	
	page_address_write=0;
	

	while(1){
		uint8_t *tmptr=tembuf+2*reads;
		rx_data = k_fifo_get(&flash_write_fifo, K_FOREVER);
		*(uint16_t*)tmptr=rx_data->counter;
		printk(" in write thread %d \n", rx_data->counter);
			if(++reads>=128){

				//printk("addr offset %x \n", FLASH_TEST_REGION_OFFSET+page_address_write*FLASH_BYTES_PER_PAGE);
				// st = k_cycle_get_32()/32;
				 rc = flash_write(flash_dev, FLASH_TEST_REGION_OFFSET+page_address_write*FLASH_BYTES_PER_PAGE, tembuf, FLASH_BYTES_PER_PAGE);
				// ed=k_cycle_get_32()/32;
				 if (rc != 0) {
				 	printf("Flash write failed! %d\n", rc);
				 	return;
				 }else{
				 	printk("page_address_write %d, count %d, write %d, %d \n",page_address_write ,count, st,ed);
				 }
				reads = 0;
				if(++page_address_write >= ANALOG_NUM_BUF_pages) {
					timer_deinit();
					printk("reach page %d\n",ANALOG_NUM_BUF_pages);
				
					
				}
			}	
		
	}
}
// K_THREAD_STACK_DEFINE(my_stack_area, 1024);
// struct k_thread my_thread_data;

// k_tid_t my_tid = k_thread_create(&my_thread_data, my_stack_area,
//                                  K_THREAD_STACK_SIZEOF(my_stack_area),
//                                  flash_write_thread,
//                                  NULL, NULL, NULL,
//                                  7, 0, K_NO_WAIT);

 K_THREAD_DEFINE(write_to_flash_id, 1024, flash_write_thread, NULL, NULL, NULL,
		7, 0, 0);
void main(void)
{	
	timer_init();
	k_thread_start(write_to_flash_id);
	while(1){
	}
		
}